@@ -49,8 +49,9 @@ public:
 	void stop();
 	/** Designates whether KIOFuse should perform FileJob-based (KIO::open) I/O where possible. */
 	void setUseFileJob(bool useFileJob);
-	/** Runs KIO::stat on url and adds a node to the tree if successful. Calls the callback at the end. */
-	void mountUrl(QUrl url, std::function<void(const std::shared_ptr<KIOFuseNode>&, int)> callback);
+	/** Runs KIO::stat on url and (if successful) creates an origin node at the lowest possible level.
+	  * Returns the relative path to where url is reachable in the callback. */
+	void mountUrl(QUrl url, std::function<void(const QString&, int)> callback);
 	/** Converts a local path into a remote URL if it is mounted within the VFS */
 	QUrl localPathToRemoteUrl(const QString &localPath) const;
 	/** Returns the path upwards until a root node. */
@@ -98,7 +99,7 @@ private:
 	static void signalHandler(int signal);
 
 	/** Returns a pointer to a child node of parent with m_nodeName == name or nullptr. */
-	std::shared_ptr<KIOFuseNode> nodeByName(const std::shared_ptr<KIOFuseNode> &parent, const QString name) const;
+	std::shared_ptr<KIOFuseNode> nodeByName(const std::shared_ptr<KIOFuseDirNode> &parent, const QString name) const;
 	/** Returns a pointer to the KIOFuseNode with inode number ino or nullptr. */
 	std::shared_ptr<KIOFuseNode> nodeForIno(const fuse_ino_t ino) const;
 	/** Removes the node from the old parent's children list (if any) and adds it to the new parent's list.*/
@@ -145,11 +146,16 @@ private:
 	void awaitNodeFlushed(const std::shared_ptr<KIOFuseRemoteCacheBasedFileNode> &node, std::function<void(int error)> callback);
 	/** Invokes callback on error or when a node has been refreshed (if its stat timed out) */
 	void awaitAttrRefreshed(const std::shared_ptr<KIOFuseNode> &node, std::function<void(int error)> callback);
+	/** Invokes callback on error on when the child node was fetched and created/updated. */
+	void awaitChildMounted(const std::shared_ptr<KIOFuseRemoteDirNode> &node, const QString name, std::function<void(const std::shared_ptr<KIOFuseNode>&, int)> callback);
 
-	/** Returns the override URL for an origin node */
-	QUrl makeOriginUrl(QUrl url);
-	/** If authority of URL is null, adds an empty authority instead */
-	QUrl sanitizeNullAuthority(QUrl url) const;
+	/** Returns the URL pointing to the origin of the linked resource, i.e. path set to / or empty. */
+	QUrl originOfUrl(QUrl url);
+	/** Returns the path elements where the URL url gets mapped to in this VFS. */
+	QStringList mapUrlToVfs(QUrl url);
+	/** Stats url. If successful, returns the path where url + pathElements is reachable in callback.
+	  * If it failed, it moves one part of pathElements to url and tries again, recursively. */
+	void findAndCreateOrigin(QUrl url, QStringList pathElements, std::function<void(const QString&, int)> callback);
     
 	/** Returns the corresponding FUSE error to the given KIO Job error */
 	static int kioErrorToFuseError(const int kioError);
@@ -163,8 +169,12 @@ private:
 
 	/** Fuse bookkeeping. */
 	struct fuse_session *m_fuseSession = nullptr;
+	/** Allow parsing of options in init() */
+	std::unique_ptr<struct fuse_conn_info_opts, decltype(&free)> m_fuseConnInfoOpts{nullptr, &free};
 	/** Fuse bookkeeping. */
 	std::unique_ptr<QSocketNotifier> m_fuseNotifier;
+	/** Path where this VFS is currently mounted at, with trailing '/'. */
+	QString m_mountpoint;
 
 	/** Fds of paired sockets. 
 	 * Used in conjunction with socket notifier to allow handling signals with the Qt event loop. **/
